import json
import os
import time
from typing import Union, Tuple, Any, Dict, List, Optional


class Option(object):
    def __init__(self, name, from_type=None, default_from=None, converter=None, default=None, arg_name=None):
        self.name = name
        self.from_type = from_type
        self.default_from = default_from
        self.converter = converter
        self.default = default
        self.arg_name = arg_name or name
        if converter is not None and default_from is not None and default is None:
            self.default = self.convert(default_from)

    def convert(self, value):
        if value is None:
            return self.default
        elif self.converter:
            if isinstance(self.from_type, tuple):
                parts = value.split(":")
                if len(parts) != len(self.from_type):
                    raise RuntimeError("Could not parse arguments for option {}, got: {}".format(self.name, value))
                value = tuple(t(v) for t, v in zip(self.from_type, parts))

            if isinstance(value, tuple):
                return self.converter(*value)
            else:
                return self.converter(value)
        else:
            return value


class Options(object):
    def __init__(self, callback=None):
        self.options = dict()
        self.values = dict()
        self.original_values = dict()
        self.callback = callback

    def add_option(self, name, from_type=None, default_from=None, converter=None, default=None, arg_name=None):
        if isinstance(default_from, tuple):
            default_from = ":".join(str(e) for e in default_from)
        self.options[name] = Option(name, from_type, default_from, converter, default, arg_name)

    def set_values(self, convert=True, **kwargs):
        for key, value in kwargs.items():
            self.set_value(key, value, convert)

    @staticmethod
    def convert_dict(**kwargs):
        def convert(*args):
            if args[0] in kwargs:
                if len(args) > 1:
                    return kwargs[args[0]](*args[1:])
                return kwargs[args[0]]
            raise RuntimeError("Unknown option {}, should be one of: {}".format(args[0], list(kwargs.keys())))

        return convert

    def set_value(self, name, value, convert=True):
        self.original_values[name] = value
        if convert:
            self.values[name] = self.options[name].convert(value)
        else:
            self.values[name] = value

    def __setattr__(self, key, value):
        if key in ["options", "values", "original_values", "callback"] or key.startswith("__"):
            return super().__setattr__(key, value)
        self.set_value(key, value)

    def __getattr__(self, item):
        if item in ["options", "values", "original_values", "callback"] or item.startswith("__"):
            return super().__getattr__(item)
        return self.values[item] if item in self.values else self.options[item].default

    def add_arguments(self, parser):
        for o_name, option in self.options.items():
            parser.add_argument(
                "--{}".format(option.name),
                type=option.from_type if not isinstance(option.from_type, tuple) else str,
                default=option.default_from
            )

    def parse_arguments(self, args):
        for o_name, option in self.options.items():
            self.set_value(option.name, getattr(args, option.name))

    def print_arguments(self):
        return " ".join("--{} {}".format(name, o_value) for name, o_value in self.original_values.items()
                        if o_value is not None)

    def call(self, timed=False) -> Union[Tuple[Any, float], Any]:
        def make_call():
            return self.callback(**{self.options[o_name].arg_name: value for o_name, value in self.values.items()})

        if timed:
            start_time = time.time()
            result = make_call()
            duration = time.time() - start_time
            return result, duration
        else:
            return make_call()

    def execute_from_command_line(self, description: str=None, timed: bool=False) -> Union[Tuple[Any, float], Any]:
        import argparse
        parser = argparse.ArgumentParser(description=description)
        self.add_arguments(parser)
        self.parse_arguments(parser.parse_args())
        return self.call(timed)

    def copy(self):
        options = self.make_copy()
        options.options = dict(self.options)
        options.values = dict(self.values)
        options.original_values = dict(self.original_values)
        return options

    def make_copy(self):
        return Options(self.callback)

    def export_to_dict(self):
        return dict(self.original_values)

    def import_from_dict(self, values_dict):
        self.set_values(True, **values_dict)


class Results(Options):
    @staticmethod
    def make_converter(converter):
        def convert(result, duration):
            return converter(result)
        return convert

    def add_result(self, name, converter):
        self.add_option(name, converter=Results.make_converter(converter))

    def add_duration(self, name="duration"):
        def convert(result, duration):
            return duration
        self.add_option(name, converter=convert)

    def export_to_dict(self):
        return dict(self.values)

    def import_from_dict(self, values_dict):
        self.set_values(False, **values_dict)


class Experiment(object):
    def __init__(self, parameters: Options, results: Options, config: Optional[Options]=None, import_handler=None):
        self.parameters = parameters
        self.results = results
        self.config = config
        self.import_handler = import_handler
        self.derived = dict()
        self.imported_from_file = None

    def register_derived(self, name, callback):
        self.derived[name] = callback

    def import_from_command_line(self):
        import argparse
        parser = argparse.ArgumentParser()
        self.parameters.add_arguments(parser)
        if self.config:
            self.config.add_arguments(parser)
        args = parser.parse_args()
        self.parameters.parse_arguments(args)
        if self.config:
            self.config.parse_arguments(args)

    def execute_from_command_line(self):
        self.import_from_command_line()
        self.execute()

    def execute(self):
        result = self.parameters.call(timed=True)
        for o_name in self.results.options:
            self.results.set_value(o_name, result)

    def export_to_dict(self):
        return {"parameters": self.parameters.export_to_dict(), "results": self.results.export_to_dict(),
                "config": self.config.export_to_dict() if self.config else None}

    def save(self, filename):
        with open(filename, "w") as ref:
            json.dump(self.export_to_dict(), ref)

    def import_from_dict(self, values_dict):
        parameters_dict, results_dict, config_dict = (values_dict[k] for k in ["parameters", "results", "config"])
        if self.import_handler is not None:
            self.import_handler(parameters_dict, results_dict, config_dict)
        self.parameters.import_from_dict(parameters_dict)
        self.results.import_from_dict(results_dict)
        if self.config and config_dict:
            self.config.import_from_dict(config_dict)

    def load(self, filename):
        with open(filename, "r") as ref:
            self.import_from_dict(json.load(ref))
        self.imported_from_file = os.path.realpath(filename)
        return self






from __future__ import print_function

import itertools
import math
import random
import time
from math import sqrt

import matplotlib as mpl
from pysmt.shortcuts import *

import plotting
from generator import get_sample
from inc_logging import LoggingObserver
from incremental_learner import AllViolationsStrategy, RandomViolationsStrategy
from k_cnf_smt_learner import KCnfSmtLearner
from k_dnf_logic_learner import KDNFLogicLearner, GreedyLogicDNFLearner, GreedyMaxRuleLearner
from k_dnf_smt_learner import KDnfSmtLearner
from k_dnf_greedy_learner import GreedyMilpRuleLearner
from parameter_free_learner import learn_bottom_up
from problem import Domain, Problem
from smt_check import SmtChecker

import parse
from smt_print import pretty_print
from virtual_data import OneClassStrategy

mpl.use('TkAgg')
import matplotlib.pyplot as plt

from k_dnf_learner import KDNFLearner

def xy_domain():
    variables = ["x", "y"]
    var_types = {"x": REAL, "y": REAL}
    var_domains = {"x": (0, 1), "y": (0, 1)}
    return Domain(variables, var_types, var_domains)


def simple_checker_problem():

    theory = Or(
        And(LE(Symbol("x", REAL), Real(0.5)), LE(Symbol("y", REAL), Real(0.5))),
        And(GT(Symbol("x", REAL), Real(0.5)), GT(Symbol("y", REAL), Real(0.5)))
    )

    return Problem(xy_domain(), theory, "simple_checker")


def simple_checker_problem_cnf():
    x, y = (Symbol(n, REAL) for n in ["x", "y"])
    theory = ((x <= 0.5) | (y > 0.5)) & ((x > 0.5) | (y <= 0.5))
    return Problem(xy_domain(), theory, "simple_cnf_checker")


def checker_problem():
    variables = ["x", "y", "a"]
    var_types = {"x": REAL, "y": REAL, "a": BOOL}
    var_domains = {"x": (0, 1), "y": (0, 1)}

    theory = Or(
        And(LE(Symbol("x", REAL), Real(0.5)), LE(Symbol("y", REAL), Real(0.5)), Symbol("a", BOOL)),
        And(GT(Symbol("x", REAL), Real(0.5)), GT(Symbol("y", REAL), Real(0.5)), Symbol("a", BOOL)),
        And(GT(Symbol("x", REAL), Real(0.5)), LE(Symbol("y", REAL), Real(0.5)), Not(Symbol("a", BOOL))),
        And(LE(Symbol("x", REAL), Real(0.5)), GT(Symbol("y", REAL), Real(0.5)), Not(Symbol("a", BOOL)))
    )

    return Problem(Domain(variables, var_types, var_domains), theory, "checker")


def simple_univariate_problem():
    variables = ["x"]
    var_types = {"x": REAL}
    var_domains = {"x": (0, 1)}

    theory = LE(Symbol("x", REAL), Real(0.6))

    return Problem(Domain(variables, var_types, var_domains), theory, "one_test")


def shared_hyperplane_problem():
    domain = xy_domain()
    x, y = (domain.get_symbol(v) for v in ["x", "y"])
    # y <= -x + 1.25
    shared1 = LE(y, Plus(Times(Real(-1.0), x), Real(1.25)))
    # y >= -x + 0.75
    shared2 = GE(y, Plus(Times(Real(-1.0), x), Real(0.75)))

    # y <= x + 0.5
    h1 = LE(y, Plus(x, Real(0.5)))
    # y >= x + 0.25
    h2 = GE(y, Plus(x, Real(0.25)))

    # y <= x - 0.25
    h3 = LE(y, Plus(x, Real(-0.25)))
    # y >= x - 0.5
    h4 = GE(y, Plus(x, Real(-0.5)))
    return Problem(domain, Or(And(shared1, shared2, h1, h2), And(shared1, shared2, h3, h4)), "shared")


def cross_problem():
    domain = xy_domain()
    x, y = (domain.get_symbol(v) for v in ["x", "y"])
    top = y <= 0.9
    middle_top = y <= 0.7
    middle_bottom = y >= 0.5
    bottom = y >= 0.1

    left = x >= 0.2
    middle_left = x >= 0.4
    middle_right = x <= 0.6
    right = x <= 0.8
    theory = (top & middle_left & middle_right & bottom) | (left & middle_top & middle_bottom & right)
    return Problem(domain, theory, "cross")


def bool_xor_problem():
    variables = ["a", "b"]
    var_types = {"a": BOOL, "b": BOOL}
    var_domains = dict()
    domain = Domain(variables, var_types, var_domains)

    a, b = (domain.get_symbol(v) for v in variables)

    theory = (a & ~b) | (~a & b)
    return Problem(domain, theory, "2xor")


def ice_cream_problem():
    variables = ["chocolate", "banana", "weekend"]
    chocolate, banana, weekend = variables
    var_types = {chocolate: REAL, banana: REAL, weekend: BOOL}
    var_domains = {chocolate: (0, 1), banana: (0, 1)}
    domain = Domain(variables, var_types, var_domains)

    chocolate, banana, weekend = (domain.get_symbol(v) for v in variables)
    theory = (chocolate < 0.650)\
             & (banana < 0.550)\
             & (chocolate + 0.7 * banana <= 0.700)\
             & (chocolate + 1.2 * banana <= 0.750)\
             & (~weekend | (chocolate + 0.7 * banana <= 0.340))

    return Problem(domain, theory, "ice_cream")


def evaluate_assignment(problem, assignment):
    # substitution = {problem.domain.get_symbol(v): val for v, val in assignment.items()}
    # return problem.theory.substitute(substitution).simplify().is_true()
    return SmtChecker(assignment).check(problem.theory)


def sample(problem, n, seed=None):
    """
    Sample n instances for the given problem
    :param problem: The problem to sample from
    :param n: The number of samples
    :param seed: An optional seed
    :return: A list containing n samples and their label (True iff the sample satisfies the theory and False otherwise)
    """
    # TODO Other distributions
    samples = []
    for i in range(n):
        instance = dict()
        for v in problem.domain.variables:
            if problem.domain.var_types[v] == REAL:
                lb, ub = problem.domain.var_domains[v]
                instance[v] = random.uniform(lb, ub)
            elif problem.domain.var_types[v] == BOOL:
                instance[v] = True if random.random() < 0.5 else False
            else:
                raise RuntimeError("Unknown variable type {} for variable {}", problem.domain.var_types[v], v)
        samples.append((instance, evaluate_assignment(problem, instance)))
    return samples


def learn(domain, data):
    return None


def draw_points(feat_x, feat_y, data, name, hyperplanes=None, truth=None):
    if truth is None:
        truth = data

    true_pos = []
    false_pos = []
    true_neg = []
    false_neg = []

    for i in range(len(data)):
        row = data[i]
        point = (float(row[0][feat_x].constant_value()), float(row[0][feat_y].constant_value()))
        if truth[i][1] and row[1]:
            true_pos.append(point)
        elif truth[i][1] and not row[1]:
            false_neg.append(point)
        elif not truth[i][1] and not row[1]:
            true_neg.append(point)
        else:
            false_pos.append(point)

    if len(true_pos) > 0: plt.scatter(*zip(*true_pos), c="green", marker="o")
    if len(false_neg) > 0: plt.scatter(*zip(*false_neg), c="red", marker="o")
    if len(true_neg) > 0: plt.scatter(*zip(*true_neg), c="green", marker="x")
    if len(false_pos) > 0: plt.scatter(*zip(*false_pos), c="red", marker="x")

    if hyperplanes is not None:
        planes = [constraint_to_hyperplane(h) for conj in hyperplanes for h in conj if h.is_le() or h.is_lt()]
        for plane in planes:
            print(plane)
            if plane[0][feat_y] == 0:
                plt.plot([plane[1], plane[1]], [0, 1])
            else:
                plt.plot([0, 1], [(plane[1] - plane[0][feat_x] * x) / plane[0][feat_y] for x in [0, 1]])

    plt.xlim([0, 1])
    plt.ylim([0, 1])

    plt.savefig("{}.png".format(name))
    plt.gcf().clear()


def constraint_to_hyperplane(constraint):
    if constraint.is_le() or constraint.is_lt():
        coefficients = dict()
        left, right = constraint.args()
        if right.is_plus():
            left, right = right, left
        if left.is_plus():
            for term in left.args():
                if term.is_times():
                    c, v = term.args()
                    coefficients[v.symbol_name()] = float(c.constant_value())
                else:
                    raise RuntimeError("Unexpected value, expected product, was {}".format(term))
        else:
            raise RuntimeError("Unexpected value, expected sum, was {}".format(left))
        return coefficients, float(right.constant_value())
    raise RuntimeError("Unexpected constraint, expected inequality, was {}".format(constraint))


def find_border_points(domain, data):
    dist = []
    border_points = set()
    for i in range(len(data)):
        dist.append([])
        for j in range(len(data)):
            dist[i].append(get_distance(domain, data[i][0], data[j][0]))

    closest_others = []
    for i in range(len(data)):
        closest_other = None
        for j in range(len(data)):
            if data[j][1] != data[i][1]:
                if closest_other is None or dist[i][j] < dist[i][closest_other]:
                    closest_other = j
        closest_others.append(closest_other)

    for i in range(len(data)):
        if dist[i][closest_others[i]] == dist[closest_others[i]][closest_others[closest_others[i]]]:
            border_points.add(i)
            border_points.add(closest_others[i])

    return border_points


def find_border_points_fast(domain, data):
    positives = [i for i in range(len(data)) if data[i][1]]
    negatives = [i for i in range(len(data)) if not data[i][1]]

    closest = dict()
    for i in positives:
        for j in negatives:
            distance = get_distance(domain, data[i][0], data[j][0])
            if i not in closest or closest[i][1] > distance:
                closest[i] = (j, distance)
            if j not in closest or closest[j][1] > distance:
                closest[j] = (i, distance)

    return [i for i in range(len(data)) if closest[i][1] == closest[closest[i][0]][1]]


def get_distance(domain, example1, example2):
    distance = 0
    for v in domain.real_vars:
        distance += (example1[v] - example2[v]) ** 2
    return sqrt(distance)


def draw_border_points(feat_x, feat_y, data, border_indices, name):
    relevant_pos = []
    irrelevant_pos = []
    relevant_neg = []
    irrelevant_neg = []

    for i in range(len(data)):
        row = data[i]
        point = (row[0][feat_x], row[0][feat_y])
        if i in border_indices and row[1]:
            relevant_pos.append(point)
        elif i in border_indices and not row[1]:
            relevant_neg.append(point)
        elif i not in border_indices and row[1]:
            irrelevant_pos.append(point)
        else:
            irrelevant_neg.append(point)

    if len(relevant_pos) > 0: plt.scatter(*zip(*relevant_pos), c="green", marker="o")
    if len(irrelevant_pos) > 0: plt.scatter(*zip(*irrelevant_pos), c="grey", marker="o")
    if len(relevant_neg) > 0: plt.scatter(*zip(*relevant_neg), c="green", marker="x")
    if len(irrelevant_neg) > 0: plt.scatter(*zip(*irrelevant_neg), c="grey", marker="x")

    plt.xlim([0, 1])
    plt.ylim([0, 1])

    plt.savefig("{}.png".format(name))
    plt.gcf().clear()


def learn_parameter_free(problem, data, seed):
    feat_x, feat_y = problem.domain.real_vars

    def learn_inc(_data, i, _k, _h):
        learner = KCnfSmtLearner(_k, _h, RandomViolationsStrategy(10))
        dir_name = "../output/{}".format(problem.name)
        img_name = "{}_{}_{}_{}_{}_{}".format(learner.name, i, _k, _h, len(data), seed)
        learner.add_observer(plotting.PlottingObserver(problem.domain, data, dir_name, img_name, feat_x, feat_y))

        initial_indices = random.sample(list(range(len(data))), 20)

        learned_theory = learner.learn(problem.domain, data, initial_indices)
        print("Learned theory:\n{}".format(pretty_print(learned_theory)))
        return learned_theory

    learn_bottom_up(data, learn_inc, 1, 1)


def learn_one_class(problem, data, seed):
    feat_x, feat_y = problem.domain.real_vars
    thresholds = {feat_x: 0.1, feat_y: 0.1}
    data = [(row, label) for row, label in data if label]

    def learn_inc(_data, i, _k, _h):
        learner = KCnfSmtLearner(_k, _h, OneClassStrategy(RandomViolationsStrategy(1), thresholds))
        dir_name = "../output/{}_one_class".format(problem.name)
        img_name = "{}_{}_{}_{}_{}_{}".format(learner.name, i, _k, _h, len(data), seed)

        learner.add_observer(plotting.PlottingObserver(problem.domain, data, dir_name, img_name, feat_x, feat_y))

        # Filter boolean versions
        # learner.add_observer(plotting.PlottingObserver(data, dir_name + "/week", img_name, "chocolate", "banana",
        #                                                condition=lambda instance, _l: not instance["weekend"]))
        # learner.add_observer(plotting.PlottingObserver(data, dir_name + "/weekend", img_name, "chocolate", "banana",
        #                                                condition=lambda instance, _l: instance["weekend"]))

        initial_indices = random.sample(list(range(len(data))), 20)

        learned_theory = learner.learn(problem.domain, data, initial_indices)
        # learned_theory = Or(*[And(*planes) for planes in hyperplane_dnf])
        print("Learned theory:\n{}".format(pretty_print(learned_theory)))
        return learned_theory

    learn_bottom_up(data, learn_inc, 1, 1, 4, 2)


def learn_one_class_sample(problem, data, seed, negative_samples=None):
    feat_x, feat_y = problem.domain.real_vars
    threshold = 0.1
    data = [(row, label) for row, label in data if label]
    negative_samples = 200 if negative_samples is None else negative_samples
    real_vars = problem.domain.real_vars
    bool_vars = problem.domain.bool_vars

    def l1(p1, p2): return max(abs(p1[r] - p2[r]) for r in real_vars) if all(p1[b] == p2[b] for b in bool_vars) else 1

    def l2(p1, p2): return math.sqrt(sum((p1[r] - p2[r])**2 for r in real_vars)) if all(p1[b] == p2[b] for b in bool_vars) else 1

    negatives = []
    while len(negatives) < negative_samples:
        neg_sample = get_sample(problem.domain)
        if all(l1(neg_sample, pos) > threshold for pos, l in data):
            negatives.append((neg_sample, False))

    data += negatives

    def learn_inc(_data, i, _k, _h):
        strategy = RandomViolationsStrategy(10)
        learner = KCnfSmtLearner(_k, _h, strategy)
        dir_name = "../output/{}_one_class_sample".format(problem.name)
        img_name = "{}_{}_{}_{}_{}_{}".format(learner.name, i, _k, _h, len(data), seed)

        learner.add_observer(plotting.PlottingObserver(problem.domain, data, dir_name, img_name, feat_x, feat_y))
        learner.add_observer(LoggingObserver(None, seed, True, strategy))

        # Filter boolean versions
        # learner.add_observer(plotting.PlottingObserver(data, dir_name + "/week", img_name, "chocolate", "banana",
        #                                                condition=lambda instance, _l: not instance["weekend"]))
        # learner.add_observer(plotting.PlottingObserver(data, dir_name + "/weekend", img_name, "chocolate", "banana",
        #                                                condition=lambda instance, _l: instance["weekend"]))

        initial_indices = random.sample(list(range(len(data))), 20)

        learned_theory = learner.learn(problem.domain, data, initial_indices)
        # learned_theory = Or(*[And(*planes) for planes in hyperplane_dnf])
        print("Learned theory:\n{}".format(pretty_print(learned_theory)))
        return learned_theory

    learn_bottom_up(data, learn_inc, 1, 1)


def main():
    n = 1000
    seed = 65
    random.seed(65)
    # problem = simple_univariate_problem()
    # problem = simple_checker_problem()
    # problem = simple_checker_problem_cnf()
    # problem = shared_hyperplane_problem()
    problem = checker_problem()
    # problem = cross_problem()
    # problem = bool_xor_problem()
    # data = [
    #     ({"x": Real(0.1), "y": Real(0.1)}, True),
    #     ({"x": Real(0.9), "y": Real(0.9)}, True),
    #     ({"x": Real(0.1), "y": Real(0.9)}, False),
    #     ({"x": Real(0.9), "y": Real(0.1)}, False),
    # ]
    # problem = ice_cream_problem()

    sample_time_start = time.time()
    data = sample(problem, n, seed=seed)
    sample_time_elapsed = time.time() - sample_time_start
    print("Computing samples took {:.2f}s".format(sample_time_elapsed))

    # learn_parameter_free(problem, data, seed)
    learn_one_class(problem, data, seed)
    exit()

    start = time.time()
    border_indices = random.sample(list(range(len(data))), 20)
    # border_indices = find_border_points_fast(problem.domain, data)
    # border_finished_time = time.time() - start
    # print("Computing border points took {:.2f}s".format(border_finished_time))
    # border_indices = list(range(len(data)))
    border_points_name = "../output/{}_{}_{}".format(problem.name, len(data), seed)
    # draw_border_points("x", "y", data, border_indices, border_points_name)
    # data_subset = [data[i] for i in sorted(list(border_indices))]

    # draw_points("x", "y", data, "data")
    # learner = OCTLearner(1, 3, 0)
    # learner = KDNFLearner(2, 2, 0, negated_hyperplanes=True)
    # learner = KDNFLearner(2, 8, 0, negated_hyperplanes=False)
    # learner = GreedyLogicDNFLearner(2, 2)
    # learner = KDnfSmtLearner(8, 2)
    # learner = GreedyMaxRuleLearner(8)
    # learner = GreedyMilpRuleLearner(4, 4)
    learner = KCnfSmtLearner(5, 5, RandomViolationsStrategy(20))

    # if isinstance(learner, KDNFLearner):
    #     learner_name = "dnf"
    # elif isinstance(learner, GreedyLogicDNFLearner):
    #     learner_name = "dnf_logic_greedy"
    # elif isinstance(learner, KDNFLogicLearner):
    #     learner_name = "dnf_logic"
    # elif isinstance(learner, GreedyMaxRuleLearner):
    #     learner_name = "dnf_greedy_one_rule"
    # elif isinstance(learner, KDnfSmtLearner):
    #     learner_name = "dnf_smt"
    # else:
    #     learner_name = "dt"

    dir_name = "../output/{}".format(problem.name)
    img_name = "{}_{}_{}".format(learner.name, len(data), seed)
    learner.add_observer(plotting.PlottingObserver(data, dir_name, img_name, "x", "y"))

    domain = problem.domain
    learned_theory = learner.learn(domain, data, border_indices)
    # learned_theory = Or(*[And(*planes) for planes in hyperplane_dnf])
    print("Learned theory:\n{}".format(parse.smt_to_nested(learned_theory)))

    # learned_problem = Problem(domain, learned_theory, problem.name)
    # learned_labels = [(a, evaluate_assignment(learned_problem, a)) for a, _ in data]

    # img_name = "../output/{}_{}_{}_{}".format(learned_problem.name, learner.name, len(data), seed)

    # def filter_data(_data, _a):
    #     return [(_f, _label) for _f, _label in _data if all(_f[v] == Bool(_a[i]) for i, v in enumerate(domain.bool_vars))]

    # print(list(enumerate(domain.bool_vars)))

    # if len(domain.real_vars) == 2:
        # for bool_assignment in itertools.product([True, False], repeat=len(domain.bool_vars)):
            # bool_str = "_".join(("" if val else "not_") + var for var, val in zip(domain.bool_vars, bool_assignment))
            # filtered_learned = filter_data(learned_labels, bool_assignment)
            # filtered_data = filter_data(data, bool_assignment)
            # print("Actual labels")
            # print("\n".join("{}\tvs {}".format(filtered_data[i][1], filtered_learned[i][1]) for i in range(len(filtered_data))))
            # draw_points("x", "y", filtered_learned, img_name + bool_str, hyperplane_dnf, truth=filtered_data)
    # draw_points("x", "y", learned_labels, img_name + "_a", hyperplane_dnf, truth=data)
    # draw_points("x", "y", learned_labels, img_name + "_not_a", hyperplane_dnf, truth=data)


if __name__ == "__main__":
    main()



from __future__ import print_function, division

import argparse
import json

import os
import random
import shutil

import re
import subprocess

import time
from bitarray import bitarray

import generator
import parse
import problem
from smt_check import test
from smt_print import pretty_print
from smt_scan import load_results, get_log_messages, dump_results
import pysmt.shortcuts as smt


def migrate_results(directory, bias=None):
    summary = os.path.join(directory, "problems.txt")
    if os.path.isfile(summary):
        with open(summary, "r") as f:
            flat = json.load(f)

        for problem_id in flat:
            for sample_size in flat[problem_id]:
                if "bias" not in flat[problem_id][sample_size]:
                    flat[problem_id][sample_size]["bias"] = "cnf" if bias is None else bias

                seed, k, h = (flat[problem_id][sample_size][v] for v in ["seed", "k", "h"])

                pattern = r'{problem_id}_{size}_{seed}_\d+_\d+.txt' \
                    .format(problem_id=problem_id, size=sample_size, seed=seed)
                for old_file in os.listdir(directory):
                    if re.match(pattern, old_file):
                        new_file = old_file[:-4] + ".learning_log.txt"
                        shutil.move(os.path.join(directory, old_file), os.path.join(directory, new_file))

        with open(summary, "w") as f:
            json.dump(flat, f)


def calculate_accuracy(domain, target_formula, learned_formula):
    # from sys import path
    # path.insert(0, "/Users/samuelkolb/Documents/PhD/wmi-pa/experiments/client")
    # from run import compute_wmi
    print("Calculate accuracy:")
    # print(pretty_print(target_formula))
    # print(pretty_print(learned_formula))

    # r0, r1 = [smt.Symbol(n, smt.REAL) for n in ["r0", "r1"]]
    # b0, b1, b2, b3 = [smt.Symbol(n, smt.BOOL) for n in ["b0", "b1", "b2", "b3"]]
    # t1 = (~(1.0 <= 0.427230115861 * r0 + 1.02084935803 * r1) | ~(1.0 <= 1.59402729715 * r0 + 0.309004054118 * r1) | ~b1)
    # t2 = (b2 | (1.0 <= 1.59402729715 * r0 + 0.309004054118 * r1) | ~b0)

    # domain = problem.Domain(["x", "y"], {"x": smt.REAL, "y": smt.REAL}, {"x": (0, 1), "y": (0, 1)})
    # x, y = smt.Symbol("x", smt.REAL), smt.Symbol("y", smt.REAL)
    # t2 = (1.0 <= 1.5 * x + 0.5 * y)
    # t2 = (2 <= 3 * x + y)
    # f = (t1 & t2)

    flat = {
        "domain": problem.export_domain(domain, False),
        "query": parse.smt_to_nested(smt.Iff(target_formula, learned_formula))
    }

    print(domain)
    print(pretty_print(target_formula))
    print(pretty_print(learned_formula))
    # accuracy = list(compute_wmi(domain, [smt.Iff(target_formula, learned_formula)]))[0]

    output = str(subprocess.check_output(["/Users/samuelkolb/Documents/PhD/wmi-pa/env/bin/python",
                                     "/Users/samuelkolb/Documents/PhD/wmi-pa/experiments/client/run.py", "-s",
                                     json.dumps(flat)]))
    accuracy = float(output.split(": ")[1])
    print(accuracy)
    return accuracy


def calculate_accuracy_approx(domain, target_formula, learned_formula, samples):
    bits_target = bitarray([test(target_formula, sample) for sample in samples])
    bits_learned = bitarray([test(learned_formula, sample) for sample in samples])
    accuracy = ((bits_target & bits_learned) | (~bits_target & ~bits_learned)).count() / len(samples)
    print(accuracy)
    return accuracy


def adapt_domain_multiple(target_problem, new_bounds):
    domain = target_problem.domain
    adapted_domain = problem.Domain(domain.variables, domain.var_types, new_bounds)
    return problem.Problem(adapted_domain, target_problem.theory, target_problem.name)


def get_problem(data_dir, problem_id):
    try:
        with open(os.path.join(data_dir, "{}.txt".format(str(problem_id)))) as f:
            import generator
            s_problem = generator.import_synthetic_data(json.load(f))
        return s_problem.synthetic_problem.theory_problem
    except IOError:
        with open(os.path.join(data_dir, "problems", "{}.txt".format(str(problem_id)))) as f:
            import generator
            theory_problem = problem.import_problem(json.load(f))

        with open(os.path.join(data_dir, "summary.json"), "r") as f:
            flat = json.load(f)
        ratio_dict = flat["ratios"]
        lookup = flat["lookup"]

        adapted_problem = adapt_domain_multiple(theory_problem, ratio_dict[lookup[problem_id]]["bounds"])

        return adapted_problem


def add_accuracy(results_dir, data_dir=None, acc_sample_size=None, recompute=False):
    results_flat = load_results(results_dir)

    for problem_id in results_flat:

        if data_dir is not None:
            theory_problem = get_problem(data_dir, problem_id)
            domain = theory_problem.domain
            target_formula = theory_problem.theory
            print(problem_id)
            print(pretty_print(target_formula))
        else:
            raise RuntimeError("Data directory missing")

        for sample_size in results_flat[problem_id]:
            config = results_flat[problem_id][sample_size]
            timed_out = config.get("time_out", False)
            if not timed_out:
                learned_formula = None
                for message in get_log_messages(results_dir, config, p_id=problem_id, samples=sample_size):
                    if message["type"] == "update":
                        learned_formula = parse.nested_to_smt(message["theory"])

                print(pretty_print(learned_formula))
                print()

                if acc_sample_size is None:
                    if recompute or "exact_accuracy" not in config:
                        config["exact_accuracy"] = calculate_accuracy(domain, target_formula, learned_formula)
                else:
                    if recompute or "approx_accuracy" not in config:
                        config["approx_accuracy"] = dict()
                    acc_dict = config["approx_accuracy"]
                    if acc_sample_size not in acc_dict:
                        acc_dict[acc_sample_size] = []
                    if len(acc_dict[acc_sample_size]) < 1:
                        seed = hash(time.time())
                        random.seed(seed)
                        samples = [generator.get_sample(domain) for _ in range(acc_sample_size)]
                        acc_dict[acc_sample_size].append({
                            "acc": calculate_accuracy_approx(domain, target_formula, learned_formula, samples),
                            "seed": seed,
                        })

    dump_results(results_flat, results_dir)


def calculate_ratio(domain, formula):
    raise NotImplementedError()


def calculate_ratio_approx(formula, samples):
    bits = bitarray([test(formula, sample) for sample in samples])
    positives = bits.count() / len(samples)
    ratio = max(positives, 1 - positives)
    print("Ratio: {}".format(ratio))
    return ratio


def add_ratio(results_dir, data_dir=None, ratio_sample_size=None, recompute=False):
    results_flat = load_results(results_dir)

    ratio_cache = dict()

    for problem_id in results_flat:
        if data_dir is not None:
            theory_problem = get_problem(data_dir, problem_id)
            domain = theory_problem.domain
            formula = theory_problem.theory
        else:
            raise RuntimeError("Data directory missing")

        seed = hash(time.time())
        random.seed(seed)
        samples = [generator.get_sample(domain) for _ in range(ratio_sample_size)]

        ratio = calculate_ratio(domain, formula) if ratio_sample_size is None else calculate_ratio_approx(formula, samples)

        for sample_size in results_flat[problem_id]:
            config = results_flat[problem_id][sample_size]

            if ratio_sample_size is None:
                if recompute or "exact_ratio" not in config:
                    config["exact_ratio"] = ratio
            else:
                if recompute or "approx_ratio" not in config:
                    config["approx_ratio"] = dict()
                ratio_dict = config["approx_ratio"]
                if ratio_sample_size not in ratio_dict:
                    ratio_dict[ratio_sample_size] = []
                if len(ratio_dict[ratio_sample_size]) < 1:
                    ratio_dict[ratio_sample_size].append({
                        "ratio": ratio,
                        "seed": seed,
                    })

    dump_results(results_flat, results_dir)


if __name__ == "__main__":
    x = smt.Symbol("x", smt.REAL)
    calculate_accuracy(problem.Domain(["x"], {"x": smt.REAL}, {"x": (0, 1)}), x <= smt.Real(0.5), x <= smt.Real(0.4))


from pysmt.shortcuts import Real
from pywmi.domain import Density


class Formula(Density):
    def __init__(self, domain, support):
        super().__init__(domain, support, Real(1))

    @classmethod
    def from_state(cls, state: dict):
        density = Density.from_state(state)
        return cls(density.domain, density.support)



import json
import StringIO
import os
from os.path import join, dirname

import sys
from fabric.api import run, env, execute, cd, local, put, get, prefix, lcd
from fabric.contrib import files


def vary_synthetic_parameter(parameter_name, values, fixed_values, learner_settings, time_out=None, samples=None,
                             exp_name=None, override=False):
    default_values = {
        "data_sets": 10,
        "bool_count": 2,
        "real_count": 2,
        "bias": "cnf",
        "k": 3,
        "literals": 4,
        "half_spaces": 7,
        "samples": 1000,
        "ratio": 90,
        "errors": 0,
    }
    for key, value in fixed_values.items():
        if key not in default_values:
            raise RuntimeError("Found unknown parameter name {}".format(key))
        default_values[key] = value

    del default_values[parameter_name]

    config = {"fixed": default_values, "vary": parameter_name, "values": values, "learner": learner_settings}
    if exp_name is None:
        exp_name = "h" + str(hash(json.dumps(config)) + sys.maxsize + 1)

    print(config)

    exp_path = join("synthetic", parameter_name, exp_name)
    local_root = dirname(dirname(__file__))
    full_gen = join(local_root, exp_path)
    full_out = join(local_root, "output", exp_path)
    full_code = join(local_root, "smtlearn")
    full_api = join(full_code, "api.py")
    full_exp = join(full_code, "experiments.py")

    # Generate
    gen_config = join(full_gen, "config.json")
    if override or not os.path.exists(gen_config):
        local("mkdir -p {}".format(full_gen))

        with open(gen_config, "w") as f:
            json.dump(config, f)

        commands = []
        for value in values:
            default_values[parameter_name] = value
            options = " ".join("--{} {}".format(name, val) for name, val in default_values.items())
            command = "python {api} generate {input}/{val} {options}" \
                .format(api=full_api, input=full_gen, val=value, options=options)
            commands.append(command)
        commands.append("wait")

        local(" & ".join(commands))

    # Learn
    out_config = join(full_out, "config.json")
    if override or not os.path.exists(out_config):
        local("mkdir -p {}".format(full_out))

        with open(out_config, "w") as f:
            json.dump(config, f)

        commands = []
        for value in values:
            options = " ".join("--{} {}".format(name, val) for name, val in learner_settings.items())
            command = "python {exp} {input}/{val} \"\" {output}/{val} {options}" \
                .format(exp=full_exp, input=full_gen, output=full_out, val=value, options=options)
            if time_out is not None:
                command += " -t {}".format(time_out)
            commands.append(command)
        commands.append("wait")

        local(" & ".join(commands))

    # Combine
    if override or not os.path.exists(join(full_out, "summary")):
        with lcd(full_gen):
            local("mkdir -p all")
            for value in values:
                local("cp {}/* all/".format(value))

        local("python {api} combine {output}/summary {values} -p {output}/"
            .format(api=full_api, output=full_out, values=" ".join(str(v) for v in values)))

        for migration in ["ratio", "accuracy"]:
            command = "python {api} migrate {migration} {output}/summary -d {input}/all" \
                .format(output=full_out, input=full_gen, values=" ".join(str(v) for v in values), api=full_api,
                        migration=migration)
            if samples is not None:
                command += " -s {}".format(samples)
            local(command)


def vary_h(time_out=None, samples=None, override=False):
    parameter = "half_spaces"
    values = [3, 4, 5, 6, 7, 8, 9, 10]
    fixed_values = {"data_sets": 100, "bool_count": 0, "real_count": 2, "k": 2, "literals": 3}

    learner = {"bias": "cnf", "selection": "random"}
    vary_synthetic_parameter(parameter, values, fixed_values, learner, time_out, samples, "standard", override)

    learner["selection"] = "dt_weighted"
    vary_synthetic_parameter(parameter, values, fixed_values, learner, time_out, samples, "dt", override)


def vary_h_simple(time_out=None, samples=None):
    parameter_name = "half_spaces"
    values = [3, 4, 5, 6, 7, 8]
    fixed_values = {"data_sets": 10, "bool_count": 0, "real_count": 2, "k": 2, "literals": 3}

    learner = {"bias": "cnf", "selection": "random"}
    vary_synthetic_parameter(parameter_name, values, fixed_values, learner, time_out, samples, "small_standard")

    learner["selection_size"] = 1
    vary_synthetic_parameter(parameter_name, values, fixed_values, learner, time_out, samples, "small_standard_single")

    learner["selection_size"] = 20
    vary_synthetic_parameter(parameter_name, values, fixed_values, learner, time_out, samples, "small_standard_20")

    learner["selection"] = "dt_weighted"
    learner["selection_size"] = 1
    vary_synthetic_parameter(parameter_name, values, fixed_values, learner, time_out, samples, "small_dt_1")

    learner["selection"] = "dt"
    learner["selection_size"] = 1
    vary_synthetic_parameter(parameter_name, values, fixed_values, learner, time_out, samples, "small_sdt_1")


if __name__ == "__main__":
    import authenticate

    authenticate.config()
    execute(vary_h_simple, time_out=200, samples=1000)



from __future__ import print_function, division

import argparse
import random

import os
from glob import glob

import numpy as np
import pysmt.shortcuts as smt
import time

from pywmi.export import Exportable
from typing import List

from incal import Formula
from incal.learner import Learner
from pywmi import Domain, evaluate
from pywmi.sample import uniform


class InsufficientBalanceError(RuntimeError):
    def __init__(self, partial_samples=None):
        self.partial_samples = partial_samples


class SyntheticFormula(Formula):
    def __init__(
        self,
        domain,
        support,
        cnf_or_dnf,
        formula_count,
        terms_per_formula,
        half_space_count,
        sample_count,
        ratio_percent,
        name=None,
        seed=None,
    ):
        super().__init__(domain, support)
        self.cnf_or_dnf = cnf_or_dnf
        self.formula_count = formula_count
        self.terms_per_formula = terms_per_formula
        self.half_space_count = half_space_count
        self.sample_count = sample_count
        self.ratio_percent = ratio_percent
        self.name = name
        self.seed = seed

    @property
    def bool_count(self):
        return len(self.domain.bool_vars)

    @property
    def real_count(self):
        return len(self.domain.real_vars)

    @property
    def k(self):
        return self.formula_count

    @property
    def h(self):
        return self.half_space_count

    @property
    def literals(self):
        return self.terms_per_formula

    def get_data(self, sample_count, max_ratio):
        samples, labels = get_problem_samples(
            self.domain, self.support, sample_count, max_ratio
        )
        return SyntheticData(self, samples, labels)

    def get_state(self):
        state = Formula.get_state(self)
        state["cnf_or_dnf"] = self.cnf_or_dnf
        state["formula_count"] = self.formula_count
        state["terms_per_formula"] = self.terms_per_formula
        state["half_space_count"] = self.half_space_count
        state["sample_count"] = self.sample_count
        state["ratio_percent"] = self.ratio_percent
        state["name"] = self.name
        state["seed"] = self.seed
        return state

    @classmethod
    def from_state(cls, state: dict):
        formula = Formula.from_state(state)
        return SyntheticFormula(
            formula.domain,
            formula.support,
            state["cnf_or_dnf"],
            state["formula_count"],
            state["terms_per_formula"],
            state["half_space_count"],
            state.get("sample_count", None),
            state.get("ratio_percent", None),
            state.get("name", None),
            state.get("seed", None),
        )


class SyntheticData(Exportable):
    def __init__(self, formula, samples, labels):
        # type: (SyntheticFormula, np.ndarray, np.ndarray) -> None
        self.formula = formula
        self.samples = samples
        self.labels = labels

    def get_state(self):
        return {
            "formula": self.formula.get_state(),
            "samples": self.samples.tolist(),
            "labels": self.labels.tolist(),
        }

    @classmethod
    def from_state(cls, state: dict):
        return cls(
            SyntheticFormula.from_state(state["formula"]),
            np.array(state["samples"]),
            np.array(state["labels"]),
        )


def generate_half_space_sample(domain, real_count):
    samples = uniform(domain, real_count)
    coefficients, offset = Learner.fit_hyperplane(domain, samples)
    coefficients = [
        smt.Real(float(coefficients[i][0])) * domain.get_symbol(domain.real_vars[i])
        for i in range(real_count)
    ]
    if random.random() < 0.5:
        return smt.Plus(*coefficients) <= offset
    else:
        return smt.Plus(*coefficients) >= offset


def generate_domain(bool_count, real_count):
    variables = ["b{}".format(i) for i in range(bool_count)] + [
        "r{}".format(i) for i in range(real_count)
    ]
    var_types = dict()
    var_domains = dict()
    for i, v in enumerate(variables):
        if i < bool_count:
            var_types[v] = smt.BOOL
        else:
            var_types[v] = smt.REAL
            var_domains[v] = (0, 1)

    return Domain(variables, var_types, var_domains)


def get_problem_samples(domain, support, sample_count, max_ratio):
    minimal_count = sample_count * min(max_ratio, 1 - max_ratio)
    samples = uniform(domain, sample_count)
    labels = evaluate(domain, support, samples)
    positive_count = sum(labels)
    if (
        positive_count < minimal_count
        or (sample_count - positive_count) < minimal_count
    ):
        raise InsufficientBalanceError()

    return samples, labels


def get_synthetic_problem_name(
    prefix,
    bool_count,
    real_count,
    cnf_or_dnf,
    k,
    l_per_term,
    h,
    sample_count,
    seed,
    ratio_percent,
    i=None,
):
    name = "{prefix}_{bc}_{rc}_{type}_{fc}_{tpf}_{hc}_{sc}_{seed}_{ratio}".format(
        bc=bool_count,
        rc=real_count,
        type=cnf_or_dnf,
        fc=k,
        tpf=l_per_term,
        hc=h,
        sc=sample_count,
        prefix=prefix,
        seed=seed,
        ratio=int(ratio_percent),
    )
    if i is not None:
        name = name + "_" + str(i)
    return name


class GeneratorError(RuntimeError):
    pass


class Generator(object):
    def __init__(
        self,
        bool_count,
        real_count,
        bias,
        k,
        l,
        h,
        sample_count,
        max_ratio,
        seed,
        prefix,
    ):
        self.domain = generate_domain(bool_count, real_count)
        self.bias = bias
        self.k = k
        self.l = l
        self.h = h
        self.sample_count = sample_count
        self.max_ratio = int(max_ratio * 100) / 100
        self.seed = seed
        self.prefix = prefix
        self.max_tries = 1000

    @property
    def bool_count(self):
        return len(self.domain.bool_vars)

    @property
    def real_count(self):
        return len(self.domain.real_vars)

    def symbol(self, var_name):
        return self.domain.get_symbol(var_name)

    def test_ratio(self, labels):
        max_ratio = max(self.max_ratio, 1 - self.max_ratio)
        return (1 - max_ratio) <= sum(labels) / self.sample_count <= max_ratio

    def get_name(self):
        b = self.bool_count
        r = self.real_count
        ratio = int(self.max_ratio * 100)
        k, l, h = self.k, self.l, self.h
        return get_synthetic_problem_name(
            self.prefix, b, r, self.bias, k, l, h, self.sample_count, self.seed, ratio
        )

    def get_samples(self):
        return uniform(self.domain, self.sample_count)

    def get_half_spaces(self, samples):
        half_spaces = []
        print("Generating half spaces: ", end="")
        if self.real_count > 0:
            while len(half_spaces) < self.h:
                half_space = generate_half_space_sample(self.domain, self.real_count)
                labels = evaluate(self.domain, half_space, samples)
                half_spaces.append((half_space, labels))
                print("y", end="")

        print()
        return half_spaces

    def get_term(self, literal_pool):
        print("Generate term: ", end="")
        for i in range(self.max_tries):
            literals = random.sample(literal_pool, self.l)

            term = smt.Or(*list(zip(*literals))[0])

            covered = np.zeros(self.sample_count)
            significant_literals = 0
            for _, labels in literals:
                prev_size = sum(covered)
                covered = np.logical_or(covered, labels)
                if (sum(covered) - prev_size) / self.sample_count >= 0.05:
                    significant_literals += 1

            if significant_literals == self.l:  # & test_ratio(covered):
                print("y", end="")
                print()
                return term, covered
            else:
                print("x", end="")
        print(" Failed after {} tries".format(self.max_tries))
        raise GeneratorError()

    def get_formula(self, name, literal_pool):
        print("Generate formula:")
        for i in range(self.max_tries):
            terms = [self.get_term(literal_pool) for _ in range(self.k)]
            formula = smt.And(*list(zip(*terms))[0])

            covered = np.ones(self.sample_count)
            significant_terms = 0
            for _, labels in terms:
                prev_size = sum(covered)
                covered = np.logical_and(covered, labels)
                if (prev_size - sum(covered)) / self.sample_count >= 0.05:
                    significant_terms += 1

            if significant_terms == self.k and self.test_ratio(covered):
                print("y({:.2f})".format(sum(covered) / self.sample_count), end="")
                return SyntheticFormula(
                    self.domain,
                    formula,
                    "cnf",
                    self.k,
                    self.l,
                    self.h,
                    self.sample_count,
                    int(self.max_ratio * 100),
                    name,
                )
            else:
                if significant_terms == self.k:
                    print("Ratio not satisfied")
                else:
                    print("Not enough significant terms")
        print("Failed to generate formula after {} tries".format(self.max_tries))
        raise GeneratorError()

    def generate_formula(self):
        for i in range(self.max_tries):
            print("Generating formula...")
            samples = self.get_samples()
            half_spaces = self.get_half_spaces(self.get_samples())
            bool_literals = [
                (self.symbol(v), samples[:, self.domain.variables.index(v)])
                for v in self.domain.bool_vars
            ]
            literal_pool = half_spaces + bool_literals
            literal_pool += [
                (smt.Not(l), np.logical_not(bits)) for l, bits in literal_pool
            ]
            print([(v, len(l)) for v, l in literal_pool])

            try:
                formula = self.get_formula(self.get_name(), literal_pool)
                return formula
            except GeneratorError:
                continue

        raise RuntimeError(
            "Could not generate a formula within {} tries".format(self.max_tries)
        )

    def generate_data_set(self):
        for i in range(self.max_tries):
            try:
                return self.generate_formula().get_data(
                    self.sample_count, self.max_ratio
                )
            except InsufficientBalanceError:
                pass
        raise RuntimeError(
            "Could not generate a dataset within {} tries".format(self.max_tries)
        )


def get_seed():
    return random.randint(0, 2 ** 32 - 1)


def generate_synthetic_formula(
    prefix: str,
    boolean_count: int,
    real_count: int,
    cnf_or_dnf: str,
    formula_count: int,
    terms_per_formula: int,
    half_space_count: int,
    sample_count: int,
    ratio_percent: int,
    seed=None,
) -> SyntheticFormula:
    seed = seed or get_seed()
    random.seed(seed)
    np.random.seed(seed)

    ratio = ratio_percent / 100
    producer = Generator(
        boolean_count,
        real_count,
        cnf_or_dnf,
        formula_count,
        terms_per_formula,
        half_space_count,
        sample_count,
        ratio,
        seed,
        prefix,
    )
    return producer.generate_formula()


def generate_synthetic_dataset(
    prefix: str,
    boolean_count: int,
    real_count: int,
    cnf_or_dnf: str,
    formula_count: int,
    terms_per_formula: int,
    half_space_count: int,
    sample_count: int,
    ratio_percent: int,
    seed=None,
) -> SyntheticData:
    seed = seed or get_seed()
    random.seed(seed)
    np.random.seed(seed)

    ratio = ratio_percent / 100
    producer = Generator(
        boolean_count,
        real_count,
        cnf_or_dnf,
        formula_count,
        terms_per_formula,
        half_space_count,
        sample_count,
        ratio,
        seed,
        prefix,
    )
    return producer.generate_data_set()


def generate_formula_and_data(
    prefix: str,
    boolean_count: int,
    real_count: int,
    cnf_or_dnf: str,
    formula_count: int,
    terms_per_formula: int,
    half_space_count: int,
    sample_count: int,
    ratio_percent: int,
    seed=None,
) -> SyntheticData:
    seed = seed or get_seed()
    random.seed(seed)
    np.random.seed(seed)

    ratio = ratio_percent / 100
    producer = Generator(
        boolean_count,
        real_count,
        cnf_or_dnf,
        formula_count,
        terms_per_formula,
        half_space_count,
        sample_count,
        ratio,
        seed,
        prefix,
    )
    return producer.generate_data_set()


def generate_formula_file(
    dir_filename: str,
    prefix: str,
    boolean_count: int,
    real_count: int,
    cnf_or_dnf: str,
    formula_count: int,
    terms_per_formula: int,
    half_space_count: int,
    sample_count: int,
    ratio_percent: int,
    seed=None,
    suffix="",
):

    seed = seed or get_seed()
    random.seed(seed)
    np.random.seed(seed)

    if len(prefix) > 0:
        prefix = f"{prefix}_"

    if len(suffix) > 0:
        suffix = f"_{suffix}"

    name = (
        f"{prefix}"
        f"{cnf_or_dnf}_"
        f"b{boolean_count}_"
        f"r{real_count}_"
        f"k{terms_per_formula}_"
        f"l{terms_per_formula}_"
        f"h{half_space_count}_"
        f"sc{sample_count}_"
        f"r{ratio_percent}_"
        f"s{seed}{suffix}"
    )

    f = generate_synthetic_formula(
        prefix,
        boolean_count,
        real_count,
        cnf_or_dnf,
        formula_count,
        terms_per_formula,
        half_space_count,
        sample_count,
        ratio_percent,
    )
    filename = os.path.join(dir_filename, f"{name}.json")
    f.to_file(filename)
    return filename


def generate_data_files(
    formula_filename,
    sample_count=None,
    seed=None,
    suffix="",
    destination_directory=None,
):
    destination_directory = destination_directory or os.path.dirname(formula_filename)
    name = os.path.basename(".".join(formula_filename.split(".")[:-1]))
    f = SyntheticFormula.from_file(formula_filename)

    sample_count = (sample_count or f.sample_count) or 1000
    seed = seed or get_seed()
    random.seed(seed)
    np.random.seed(seed)
    suffix = suffix if len(suffix) == 0 else f"_{suffix}"

    data = uniform(f.domain, 1000)
    data_filename = os.path.join(
        destination_directory, f"{name}_dsc{sample_count}_ds{seed}{suffix}.data"
    )
    np.save(data_filename, data)

    labels = evaluate(f.domain, f.support, data)
    labels_filename = os.path.join(
        destination_directory, f"{name}_dsc{sample_count}_ds{seed}{suffix}.labels"
    )
    np.save(labels_filename, labels)

    return data_filename, labels_filename


def clean_dir(dir_filename, prefix=None):
    dir_filename = os.path.abspath(dir_filename)
    prefix = prefix or ""
    for f in glob(os.path.join(dir_filename, f"{prefix}*.json")):
        os.unlink(f)
    for f in glob(os.path.join(dir_filename, f"{prefix}*.npy")):
        os.unlink(f)


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument("data_dir")
    parser.add_argument("--problems", default=None, type=int)
    parser.add_argument("--datasets", default=None, type=int)
    parser.add_argument("--prefix", default="synthetic")
    parser.add_argument("--bool_count", default=2, type=int)
    parser.add_argument("--real_count", default=2, type=int)
    parser.add_argument("--bias", default="cnf")
    parser.add_argument("--k", default=3, type=int)
    parser.add_argument("--literals", default=4, type=int)
    parser.add_argument("--h", default=7, type=int)
    parser.add_argument("--sample_count", default=1000, type=int)
    parser.add_argument("--ratio", default=90, type=int)
    parsed = parser.parse_args()

    data_dir = parsed.data_dir
    print(data_dir)
    if not os.path.exists(data_dir):
        os.makedirs(data_dir)

    if parsed.problems:
        for _ in range(parsed.problems):
            formula = generate_synthetic_formula(
                parsed.prefix,
                parsed.bool_count,
                parsed.real_count,
                parsed.bias,
                parsed.k,
                parsed.literals,
                parsed.h,
                parsed.sample_count,
                parsed.ratio,
            )
            formula.to_file(os.path.join(data_dir, formula.name + ".json"))

    if parsed.datasets:
        for _ in range(parsed.datasets):
            dataset = generate_synthetic_dataset(
                parsed.prefix,
                parsed.bool_count,
                parsed.real_count,
                parsed.bias,
                parsed.k,
                parsed.literals,
                parsed.h,
                parsed.sample_count,
                parsed.ratio,
            )
            dataset.to_file(os.path.join(data_dir, dataset.formula.name + ".data.json"))
